(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('glob'), require('child_process'), require('koishi-core'), require('koishi-utils'), require('@koa/router'), require('koa-logger'), require('path'), require('fs'), require('axios')) :
  typeof define === 'function' && define.amd ? define(['exports', 'glob', 'child_process', 'koishi-core', 'koishi-utils', '@koa/router', 'koa-logger', 'path', 'fs', 'axios'], factory) :
  (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global['koishi-plugin-manager'] = {}, global.glob, global.child_process, global.koishiCore, global.koishiUtils, global.Router, global.KoaLogger, global.path, global.fs, global.axios));
}(this, (function (exports, glob, child_process, koishiCore, koishiUtils, Router, KoaLogger, path, fs, axios) { 'use strict';

  function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

  var glob__default = /*#__PURE__*/_interopDefaultLegacy(glob);
  var Router__default = /*#__PURE__*/_interopDefaultLegacy(Router);
  var KoaLogger__default = /*#__PURE__*/_interopDefaultLegacy(KoaLogger);
  var path__default = /*#__PURE__*/_interopDefaultLegacy(path);
  var fs__default = /*#__PURE__*/_interopDefaultLegacy(fs);
  var axios__default = /*#__PURE__*/_interopDefaultLegacy(axios);

  const CtxProto = koishiCore.Context.prototype;
  const sourceContextMethods = {
    select: CtxProto.select,
    plugin: CtxProto.plugin,
    dispose: CtxProto.dispose
  };
  const allContexts = new Map();

  CtxProto.select = function (key, ...values) {
    const childDatas = allContexts.get(this) ?? [];
    if (childDatas.length === 0) allContexts.set(this, childDatas);
    const childContexts = childDatas.filter(childData => key === childData.key && values.toString() === childData.values.toString()).map(childData => childData.childCtx);

    if (childContexts.length >= 1) {
      return childContexts[0];
    }

    childDatas.push({
      key,
      values,
      childCtx: sourceContextMethods.select.call(this, key, ...values)
    });
    return childDatas[childDatas.length - 1].childCtx;
  };

  class AllPlugins extends Map {
    get plugins() {
      const plugins = [];
      this.forEach(pluginDatas => {
        plugins.push(...pluginDatas.map(pluginData => pluginData.plugin));
      });
      return plugins;
    }

    searchCtx(plugin) {
      let ctx = null;
      this.forEach((pluginDatas, val) => {
        if (pluginDatas.filter(pluginData => pluginData.plugin === plugin)[0] ?? null) {
          ctx = val;
        }
      });
      return ctx;
    }

  }

  const allPlugins = new AllPlugins();

  CtxProto.plugin = function (plugin, options) {
    const ctxPlugins = allPlugins.get(this) ?? [];
    if (ctxPlugins.length === 0) allPlugins.set(this, ctxPlugins);

    if (typeof plugin !== 'function' && 'name' in plugin) {
      plugin = { ...plugin
      };
      ctxPlugins.push({
        plugin,
        options
      });
    }

    return sourceContextMethods.plugin.call(this, plugin, options);
  };

  CtxProto.dispose = async function (plugin) {
    const ctxPlugins = allPlugins.get(allPlugins.searchCtx(plugin));

    if (ctxPlugins) {
      const index = ctxPlugins.findIndex(val => val.plugin.apply === plugin.apply);

      if (index >= 0) {
        ctxPlugins.splice(index, 1);
      }
    }

    await sourceContextMethods.dispose.call(this, plugin);
  };

  /**
   * @desc   Session.ts
   * @author yijie
   * @date   2021-05-19 10:26
   * @notes  2021-05-19 10:26 yijie 创建了 Session.ts 文件
   */

  koishiCore.Session.prototype.genSessionCtx = function (ctx, options) {
    let sessionCtx = ctx;

    if (!options.channel && !options.global) {
      sessionCtx = ctx.select('userId', this.userId);
    }

    if (options.channel) {
      if (!this.groupId) {
        throw new Error('当前会话不是频道，无法使用 `group` 参数。');
      }

      sessionCtx = sessionCtx.select('groupId', this.groupId);
    }

    if (options.global) {
      sessionCtx = ctx.app;
    }

    return sessionCtx;
  };

  function createCommonjsModule(fn, module) {
  	return module = { exports: {} }, fn(module, module.exports), module.exports;
  }

  // Copyright Joyent, Inc. and other Node contributors.

  // If obj.hasOwnProperty has been overridden, then calling
  // obj.hasOwnProperty(prop) will break.
  // See: https://github.com/joyent/node/issues/1707
  function hasOwnProperty(obj, prop) {
    return Object.prototype.hasOwnProperty.call(obj, prop);
  }

  var decode = function(qs, sep, eq, options) {
    sep = sep || '&';
    eq = eq || '=';
    var obj = {};

    if (typeof qs !== 'string' || qs.length === 0) {
      return obj;
    }

    var regexp = /\+/g;
    qs = qs.split(sep);

    var maxKeys = 1000;
    if (options && typeof options.maxKeys === 'number') {
      maxKeys = options.maxKeys;
    }

    var len = qs.length;
    // maxKeys <= 0 means that we should not limit keys count
    if (maxKeys > 0 && len > maxKeys) {
      len = maxKeys;
    }

    for (var i = 0; i < len; ++i) {
      var x = qs[i].replace(regexp, '%20'),
          idx = x.indexOf(eq),
          kstr, vstr, k, v;

      if (idx >= 0) {
        kstr = x.substr(0, idx);
        vstr = x.substr(idx + 1);
      } else {
        kstr = x;
        vstr = '';
      }

      k = decodeURIComponent(kstr);
      v = decodeURIComponent(vstr);

      if (!hasOwnProperty(obj, k)) {
        obj[k] = v;
      } else if (Array.isArray(obj[k])) {
        obj[k].push(v);
      } else {
        obj[k] = [obj[k], v];
      }
    }

    return obj;
  };

  // Copyright Joyent, Inc. and other Node contributors.

  var stringifyPrimitive = function(v) {
    switch (typeof v) {
      case 'string':
        return v;

      case 'boolean':
        return v ? 'true' : 'false';

      case 'number':
        return isFinite(v) ? v : '';

      default:
        return '';
    }
  };

  var encode = function(obj, sep, eq, name) {
    sep = sep || '&';
    eq = eq || '=';
    if (obj === null) {
      obj = undefined;
    }

    if (typeof obj === 'object') {
      return Object.keys(obj).map(function(k) {
        var ks = encodeURIComponent(stringifyPrimitive(k)) + eq;
        if (Array.isArray(obj[k])) {
          return obj[k].map(function(v) {
            return ks + encodeURIComponent(stringifyPrimitive(v));
          }).join(sep);
        } else {
          return ks + encodeURIComponent(stringifyPrimitive(obj[k]));
        }
      }).join(sep);

    }

    if (!name) return '';
    return encodeURIComponent(stringifyPrimitive(name)) + eq +
           encodeURIComponent(stringifyPrimitive(obj));
  };

  var querystring = createCommonjsModule(function (module, exports) {

  exports.decode = exports.parse = decode;
  exports.encode = exports.stringify = encode;
  });
  querystring.decode;
  querystring.parse;
  querystring.encode;
  var querystring_4 = querystring.stringify;

  process.env.no_proxy = process.env?.no_proxy ?? '';
  process.env.no_proxy += ',api.npms.io';

  class NpmApi {
    constructor() {
      this.axios = axios__default['default'].create({
        baseURL: 'https://api.npms.io/v2/'
      });
      this.axios.interceptors.response.use(response => {
        return response.data;
      }, error => {
        return Promise.reject(error);
      });
    }

    async search(q, from = 0, size = 10) {
      return await this.axios.get(`search?${querystring_4({
      q,
      from,
      size
    })}`);
    }

    async get(name) {
      return await this.axios.get(`package/${name}`);
    }

  }

  const npmApi = new NpmApi();

  const pluginService = {
    ctx: null,

    listFromLocal(root = process.cwd(), keyWords = []) {
      const pluginPaths = glob__default['default'].sync(path__default['default'].resolve(root, './node_modules/*koishi-plugin-*'));
      const pRoot = path__default['default'].resolve(root, '../');
      const plugins = pluginPaths.filter(pluginPath => keyWords.length === 0 || keyWords.filter(k => pluginPath.indexOf(k) >= 0).length > 0).map(pluginPath => {
        const absPath = path__default['default'].resolve(pluginPath, './package.json');
        return JSON.parse(fs__default['default'].readFileSync(absPath).toString());
      });
      (pRoot === root ? [] : this.listFromLocal(pRoot, keyWords)).forEach(pkg => {
        const index = plugins.findIndex(plugin => plugin.name === pkg.name);
        if (index == -1) plugins.push(pkg);
      });
      return plugins;
    },

    /**
     * install plugins to session
     *
     * @param plugins { string[] }     - wait install plugins
     * @param key     { K }            - select session type
     * @param values  { Session[K][] } - select value of session
     */
    async installToSession(plugins, key, values) {
      const sessionCtx = key ? this.ctx.select(key, ...values) : this.ctx.app;
      const ctxPlugins = allPlugins.get(sessionCtx);
      const waitInstalledPlugins = [];

      for (let i = 0; i < plugins.length; i++) {
        const pluginName = '' + plugins[i];
        const data = searchPlugin(pluginName);

        if (data !== null) {
          const isInstalled = ctxPlugins && ctxPlugins.filter(ctxPluginData => ctxPluginData.plugin?.apply && ctxPluginData.plugin?.apply === data.pluginModule?.apply).length >= 1;

          if (isInstalled) {
            throw new Error(`当前会话已安装 ${pluginName}`);
          } else {
            waitInstalledPlugins.push(data);
          }
        } else {
          throw new Error(`本地未安装 ${pluginName} / koishi-plugin-${pluginName}`);
        }
      }

      for (let i = 0; i < waitInstalledPlugins.length; i++) {
        const data = waitInstalledPlugins[i];

        if (key === undefined) {
          const plugins = allPlugins.plugins.filter(p => p.apply && p.apply === data.pluginModule.apply);

          for (let j = 0; j < plugins.length; j++) {
            await sessionCtx.dispose(plugins[j]);
          }
        }

        sessionCtx.plugin(data.pluginModule);
      }

      return `${waitInstalledPlugins.map(i => i.pluginName).join(', ')} 安装完成`;
    },

    async installFromRemote(plugins) {
      const localPkgs = pluginService.listFromLocal();
      const waitInstallPlugins = [];

      for (let i = 0; i < plugins.length; i++) {
        const pluginName = plugins[i];

        try {
          if (localPkgs.findIndex(pkg => pkg.name === pluginName) == -1) {
            const pkgData = await npmApi.get(pluginName);
            const pkg = pkgData.collected.metadata;
            waitInstallPlugins.push(pkg.name);
          }
        } catch (e) {
          if (e.message === 'Request failed with status code 404') {
            throw new Error(`远程不存在 ${pluginName}。`);
          }
        }
      }

      if (waitInstallPlugins.length === 0) return [];
      const args = [];
      const absPath = path__default['default'].resolve(process.cwd(), './package.json');
      const pkg = JSON.parse(fs__default['default'].readFileSync(absPath).toString());

      if (pkg.workspaces) {
        args.push('-W');
      }

      await doCommand('yarn', ['add', ...args, ...waitInstallPlugins]);
      return waitInstallPlugins;
    }

  };

  const registerInstallCmd = (ctx, cmd, logger) => {
    const checkPlugins = (_argv, ...plugins) => {
      if (plugins.length === 0) return '请输入待安装的插件';
    };

    cmd.subcommand('.install [...plugins]', '安装插件').usage('安装插件到当前会话').alias(...['i', 'ins', 'add'].map(i => `kpm.${i}`)).option('channel', '-c 频道', {
      type: 'boolean'
    }).option('global', '-g 全局', {
      type: 'boolean'
    }).check(checkPlugins).action(async ({
      session,
      options
    }, ...plugins) => {
      let key;
      const values = [];

      if (options.channel) {
        if (!session.groupId) throw new Error('当前会话不是频道，无法使用 `group` 参数。');
        key = 'groupId';
        values.push(session.groupId);
      } else {
        if (!options.global) {
          key = 'userId';
          values.push(session.userId);
        }
      }

      for (let i = 0; i < plugins.length; i++) {
        const msg = await pluginService.installToSession([plugins[i]], key, values);
        logger.info(msg);
        await session.send(msg);
      }
    }).subcommand('.remote [...plugins]', '从远程安装插件(|依赖)').check(checkPlugins).action(async ({
      session
    }, ...plugins) => {
      const pluginsStr = plugins.join(', ');

      try {
        await session.send(`${pluginsStr} 安装中`);
        const installedPlugins = await pluginService.installFromRemote(plugins);

        if (installedPlugins.length === 0) {
          return '无依赖需要安装';
        } else {
          return `${installedPlugins.join(', ')} 安装完毕.`;
        }
      } catch (e) {
        await session.send(`${pluginsStr} 安装失败.`); // 安全起见，报错信息只私聊发送给用户

        await session.bot.sendPrivateMessage(session.userId, `${new Date().toLocaleString()} 安装 ${pluginsStr} 失败\n` + `失败原因:\n ${e}`);
      }
    });
  };

  const registerUnInstallCmd = (ctx, cmd) => {
    cmd.subcommand('.uninstall [...plugins] 卸载插件').usage('从指定会话卸载插件').alias(...['uni', 'un', 'unlink', 'remove', 'rm', 'r'].map(i => `kpm.${i}`)).option('channel', '-c 频道', {
      type: 'boolean'
    }).option('global', '-g 全局', {
      type: 'boolean'
    }).action(async ({
      session,
      options
    }, ...plugins) => {
      const sessionCtx = session.genSessionCtx(ctx, options);

      for (let i = 0; i < plugins.length; i++) {
        const pluginName = '' + plugins[i];
        const data = searchPlugin(pluginName);

        if (data !== null) {
          let plugins;

          if (options.global) {
            plugins = allPlugins.plugins;
          } else {
            plugins = allPlugins.get(sessionCtx).map(pluginData => pluginData.plugin);
          }

          plugins = plugins.filter(p => p.apply && p.apply === data.pluginModule.apply);

          for (let j = 0; j < plugins.length; j++) {
            await ctx.dispose(plugins[j]);
            await session.send(`uninstalled ${plugins[j].name}`);
          }
        } else {
          await session.send(`本地未安装 ${pluginName} / koishi-plugin-${pluginName}`);
        }
      }

      return '卸载完成';
    });
  };

  const registerListCmd = (ctx, cmd) => {
    cmd.subcommand('.list', '插件列表').usage('展示当前会话已安转的插件').alias(...['ls', 'l'].map(i => `kpm.${i}`)).option('channel', '-c 频道', {
      type: 'boolean'
    }).option('global', '-g 全局', {
      type: 'boolean'
    }).action(async ({
      session,
      options
    }) => {
      const sessionCtx = session.genSessionCtx(ctx, options);
      let pluginsList = '';
      const ctxPlugins = allPlugins.get(sessionCtx);
      ctxPlugins && ctxPlugins.forEach(ctxPluginData => {
        const plugin = ctxPluginData.plugin;
        pluginsList += `[√] ${plugin?.name ?? '未命名插件'}\n`;
      });
      return pluginsList === '' ? '暂无已安装的插件' : pluginsList;
    }).subcommand('.local', '展示本地插件').alias(...['l'].map(i => `kpm.l.${i}`), ...['l'].map(i => `kpm.ls.${i}`), ...['l'].map(i => `kpm.list.${i}`)).action(() => {
      const pluginPkgs = pluginService.listFromLocal();
      let returnMsg = `本地共检索到: ${pluginPkgs.length}个依赖\n`;
      pluginPkgs.forEach(pkg => {
        returnMsg += `${pkg.name}[${pkg.author}]\n`;
        returnMsg += pkg.description + '\n';
        returnMsg += `最新版本: ${pkg.version}\n`;
        returnMsg += '='.repeat(24) + '\n';
      });
      return returnMsg;
    }).parent.subcommand('.remote [query]', '搜索远程插件(|依赖)').alias(...['r'].map(i => `kpm.l.${i}`), ...['r'].map(i => `kpm.ls.${i}`), ...['r'].map(i => `kpm.list.${i}`)).option('page', '-p [page] 当前页码', {
      type: 'number',
      fallback: 0
    }).option('size', '-s [size] 当前个数', {
      type: 'number',
      fallback: 10
    }).check(({
      options
    }) => options.size && !Number.isInteger(options.size) || options.page && !Number.isInteger(options.page) ? '参数错误' : undefined).action(async ({
      options
    }, query = 'koishi-plugin') => {
      const pluginsPagination = await npmApi.search(query, options.page, options.size);
      let returnMsg = `远程共检索到: ${pluginsPagination.total}个依赖\n`;
      pluginsPagination.results.forEach(item => {
        const pkg = item.package;
        returnMsg += `${pkg.name}[${pkg.author.name}]\n`;
        returnMsg += pkg.description + '\n';
        returnMsg += `最新版本: ${pkg.version}\n`;
        returnMsg += '='.repeat(24) + '\n';
      });
      returnMsg += `当前位于 ${options.page} 页，共 ${Math.ceil(pluginsPagination.total / options.size)} 页`;
      return returnMsg;
    });
  };

  const registerUpgradeCmd = (ctx, cmd) => {
    const checkPlugin = (_argv, plugin) => {
      if (!plugin) return '请输入需要安装的插件';
    };

    cmd.subcommand('.upgrade <plugin>', '更新插件').usage('更新插件').alias(...['up'].map(i => `kpm.${i}`)).option('channel', '-c 频道', {
      type: 'boolean'
    }).option('global', '-g 全局', {
      type: 'boolean'
    }).check(checkPlugin).action(async ({
      session,
      options
    }, plugin) => {
      const args = [];
      options.global && args.push('-g');
      options.channel && args.push('-c');
      await session.execute(`kpm.uni ${args.join(' ')} ${plugin}`);
      await session.execute(`kpm.ins ${args.join(' ')} ${plugin}`);
      return '重载插件成功';
    }).subcommand('.remote <plugin> [version]', '远程更新插件').option('scoped', '-S <scope>', {
      type: 'string'
    }).usage('通过更新依赖的方式更新插件').check(checkPlugin).action(async ({
      session,
      options
    }, plugin, version) => {
      const p = searchPlugin(plugin);

      if (p === null) {
        return `本地未安装该插件，请使用 \`kpm.ins.remote ${plugin}\` 从远程服务器安装`;
      }

      await session.send('更新将从全局卸载该插件，请确认是否继续');

      if (['no', 'n', '否', '不要', '不确认', '不'].includes(await session.prompt(10000))) {
        return '更新结束';
      } // uninstall plugin in global


      await session.execute(`kpm.uni -g ${plugin}`);

      try {
        plugin = version ? `${plugin}@${version}` : plugin;
        const argv = [];

        if (options.scoped) {
          argv.push('-S', options.scoped);
        }

        await doCommand('yarn', ['upgrade', plugin, ...argv]);
        return `${plugin} 更新完毕`;
      } catch (e) {
        await session.send(`${plugin} 更新失败.`); // 安全起见，报错信息只私聊发送给用户

        await session.bot.sendPrivateMessage(session.userId, `${new Date().toLocaleString()} 更新 ${plugin} 失败\n` + `失败原因:\n ${e}`);
      }
    });
  };

  const name = 'manager'; // 插件默认配置

  const defaultConfig = {
    restfulApi: false
  };
  const searchPlugin = plugin => {
    const pluginName = plugin;

    try {
      return {
        pluginName,
        pluginModule: require(pluginName)
      };
    } catch (e) {
      if (/^koishi-plugin-.*$/.test(pluginName)) return null;
      return searchPlugin('koishi-plugin-' + pluginName);
    }
  };
  const doCommand = (cmd, args, options) => new Promise((resolve, reject) => {
    if (process.platform === 'win32') {
      cmd += '.cmd';
    }

    const execCmd = child_process.spawn(cmd, args, options);
    let outputStr = '';
    execCmd.stdout.on('data', data => outputStr += data);
    execCmd.stderr.on('data', data => outputStr += data);
    execCmd.on('exit', code => {
      if (code === 0) {
        resolve(outputStr);
      } else {
        reject(outputStr);
      }
    });
  });
  const apply = (ctx, config = {}) => {
    const logger = ctx.logger(`koishi-plugin-${name}`);
    config = koishiUtils.merge(config, defaultConfig);
    pluginService.ctx = ctx;
    const kpmCmd = ctx.command('kpm', '插件管理工具', {
      authority: 4
    });

    kpmCmd.subcommand = function hackSubcommand(def, ...args) {
      const subcommand = Object.getPrototypeOf(this).subcommand.call(this, def, args);

      subcommand.action = function hackAction(callback, append = false) {
        return Object.getPrototypeOf(this).action.call(this, async (...args) => {
          try {
            return await callback.call(this, ...args);
          } catch (e) {
            return e.message;
          }
        }, append);
      };

      subcommand.subcommand = hackSubcommand;
      return subcommand;
    };

    registerInstallCmd(ctx, kpmCmd, logger);
    registerUnInstallCmd(ctx, kpmCmd);
    registerListCmd(ctx, kpmCmd);
    registerUpgradeCmd(ctx, kpmCmd);

    if (config.restfulApi) {
      const pluginRouter = new Router__default['default']({
        prefix: `/plugin-apis/${name}`
      });
      pluginRouter[koishiCore.Context.current] = ctx;
      pluginRouter.all('/(.*)', KoaLogger__default['default'](str => logger.info(str)));
      pluginRouter.all('/(.*)', async (koaCtx, next) => {
        try {
          await next();
        } catch (e) {
          const re = /^\[(.*)]:(.*)$/;

          if (e instanceof Error && typeof e.message === 'string' && re.test(e.message)) {
            const [_allMsg, status, msg] = re[Symbol.match](e.message);

            if (status && msg) {
              koaCtx.status = +status;
              koaCtx.body = msg;
            } else {
              throw e;
            }

            logger.error(e);
          } else {
            throw e;
          }
        }
      });

      (async () => {
        const controllerModules = await Promise.all(glob__default['default'].sync(`${path__default['default'].resolve(__dirname, './controllers')}/*.ts`).map(c => require(c)));
        controllerModules.forEach(controllerModule => {
          const r = controllerModule.router(ctx);
          pluginRouter.use(r.routes(), r.allowedMethods());
        });
        ctx.router.use(pluginRouter.routes(), pluginRouter.allowedMethods());
      })();
    }
  };

  exports.apply = apply;
  exports.doCommand = doCommand;
  exports.name = name;
  exports.searchPlugin = searchPlugin;

  Object.defineProperty(exports, '__esModule', { value: true });

})));
